from typing import Callable, List
from itertools import combinations, permutations

from mapel.core.matchings import *
from mapel.marriages.objects.Marriages import Marriages
from mapel.core.inner_distances import swap_distance


# MAIN DISTANCES
def compute_retrospective_distance(instance_1, instance_2, inner_distance):

    results = []
    for crossing in [False, True]:
        cost_table_1, cost_table_2 = get_matching_cost_retrospective(
            instance_1, instance_2, inner_distance, crossing=crossing)

        a, _ = solve_matching_vectors(cost_table_1)
        b, _ = solve_matching_vectors(cost_table_2)
        results.append(a+b)

    return min(results), None


def compute_positionwise_distance(instance_1, instance_2, inner_distance):
    cost_table = get_matching_cost_positionwise(instance_1, instance_2, inner_distance)
    return solve_matching_vectors(cost_table), None


def compute_pos_swap_distance(instance_1: Marriages, instance_2: Marriages,
                              inner_distance: Callable) -> (float, list):
    """ Compute Positionwise distance between ordinal elections """
    cost_table = get_matching_cost_positionwise(instance_1, instance_2, inner_distance)
    obj_val, matching = solve_matching_vectors(cost_table)
    votes_1 = instance_1.votes
    votes_2 = instance_2.votes
    size = instance_1.num_agents
    return sum([swap_distance(votes_1[i], votes_2[matching[i]], matching=matching) for i in range(size)])


def compute_swap_bf_distance(instance_1: Marriages, instance_2: Marriages,
                              inner_distance: Callable) -> int:

    obj_values = []
    tmp = True
    for matching in permutations(range(instance_1.num_agents)):
        votes_1 = instance_1.votes
        votes_2 = instance_2.votes
        size = instance_1.num_agents
        value = sum([swap_distance(votes_1[matching[i]], votes_2[i], matching=matching) for i in
                    range(size)])
        obj_values.append(value)
        if tmp:
            print(matching, value)
            tmp = False

    return min(obj_values)


def compute_pairwise_distance(instance_1: Marriages, instance_2: Marriages,
                              inner_distance: Callable) -> float:
    """ Compute Pairwise distance between ordinal elections """
    length = instance_1.num_agents
    matrix_1 = instance_1.votes_to_pairwise_matrix()
    matrix_2 = instance_2.votes_to_pairwise_matrix()
    return solve_matching_matrices(matrix_1, matrix_2, length, inner_distance)

# HELPER FUNCTIONS #
def print_matrix_for_tex(matrix):
    for row in matrix:
        line = ''
        for value in row[0:-1]:
            line += f' {value} &'
        line += f' {row[-1]} \\\\'
        print(line)


def get_matching_cost_retrospective(instance_1: Marriages, instance_2: Marriages,
                                    inner_distance: Callable, crossing=False):
    """ Return: Cost table """
    vectors_1 = instance_1.get_retrospective_vectors()
    vectors_2 = instance_2.get_retrospective_vectors()

    size = instance_1.num_agents

    if crossing:
        return [[inner_distance(vectors_1[0][i], vectors_2[1][j]) for i in range(size)] for j in range(size)], \
               [[inner_distance(vectors_1[1][i], vectors_2[0][j]) for i in range(size)] for j in range(size)]

    else:
        return [[inner_distance(vectors_1[0][i], vectors_2[0][j]) for i in range(size)] for j in range(size)], \
               [[inner_distance(vectors_1[1][i], vectors_2[1][j]) for i in range(size)] for j in range(size)]



def get_matching_cost_positionwise(instance_1: Marriages, instance_2: Marriages,
                                    inner_distance: Callable) -> List[list]:
    """ Return: Cost table """
    vectors_1 = instance_1.get_positionwise_vectors()
    vectors_2 = instance_2.get_positionwise_vectors()
    size = instance_1.num_agents
    return [[inner_distance(vectors_1[i], vectors_2[j]) for i in range(size)] for j in range(size)]

# def get_matching_cost_pos_swap(instance_1: Roommates, instance_2: Roommates,
#                                 matching) -> List[list]:
#     """ Return: Cost table """
#     votes_1 = instance_1.votes
#     votes_2 = instance_2.votes
#
#     size = instance_1.num_agents
#     return [[swap_distance(votes_1[i], votes_2[j], matching=matching) for i in range(size)]
#             for j in range(size)]

# # # # # # # # # # # # # # # #
# LAST CLEANUP ON: 22.10.2021 #
# # # # # # # # # # # # # # # #



#!/usr/bin/env python
import logging
from typing import Union

import numpy as np

import mapel.marriages.cultures.euclidean as euclidean
import mapel.marriages.cultures.impartial as impartial
import mapel.marriages.cultures.mallows as mallows
import mapel.marriages.cultures.urn as urn


def generate_votes(culture_id: str = None,
                   num_agents: int = None,
                   params: dict = None) -> Union[list, np.ndarray]:
    independent_models = {
        'ic': impartial.generate_ic_votes,
        'id': impartial.generate_id_votes,
        'symmetric': impartial.generate_symmetric_votes,
        'norm-mallows': mallows.generate_norm_mallows_votes,
        'urn': urn.generate_urn_votes,
        'group_ic': impartial.generate_group_ic_votes,
    }
    dependent_models = {
        'malasym': mallows.generate_mallows_asymmetric_votes,
        'asymmetric': impartial.generate_asymmetric_votes,
        'euclidean': euclidean.generate_euclidean_votes,
        'reverse_euclidean': euclidean.generate_reverse_euclidean_votes,
        'mallows_euclidean': euclidean.generate_mallows_euclidean_votes,
        'expectation': euclidean.generate_expectation_votes,
        'attributes': euclidean.generate_attributes_votes,
        'fame': euclidean.generate_fame_votes,
    }

    if culture_id in independent_models:
        votes_1 = independent_models.get(culture_id)(num_agents=num_agents, **params)
        votes_2 = independent_models.get(culture_id)(num_agents=num_agents, **params)
        return [votes_1, votes_2]

    elif culture_id in dependent_models:
        return dependent_models.get(culture_id)(num_agents=num_agents, **params)

    else:
        logging.warning(f'No such culture id: {culture_id}')
        return []

# # # # # # # # # # # # # # # #
# LAST CLEANUP ON: 22.10.2021 #
# # # # # # # # # # # # # # # #




from mapel.marriages.cultures.impartial import generate_asymmetric_votes
import mapel.core.features.mallows as ml


def generate_mallows_votes(*args, **kwargs):
    return ml.generate_mallows_votes(*args, **kwargs)


def generate_norm_mallows_votes(num_agents=None,
                                normphi=0.5,
                                weight=0.,
                                **kwargs):

    phi = ml.phi_from_normphi(num_agents, normphi=normphi)

    return generate_mallows_votes(num_agents, num_agents, phi)


def generate_mallows_asymmetric_votes(num_agents: int = None,
                                      phi: float = 0.5,
                                      **kwargs):
    """ Mallows on top of Asymmetric instance """

    votes_left, votes_right = generate_asymmetric_votes(num_agents=num_agents)

    votes_left = mallows_votes(votes_left, phi)
    votes_right = mallows_votes(votes_right, phi)

    return [votes_left, votes_right]


def mallows_vote(vote, phi):
    num_candidates = len(vote)
    raw_vote = generate_mallows_votes(1, num_candidates, phi=phi, weight=0)[0]
    new_vote = [0] * len(vote)
    for i in range(num_candidates):
        new_vote[raw_vote[i]] = vote[i]
    return new_vote


def mallows_votes(votes, phi):
    for i in range(len(votes)):
        votes[i] = mallows_vote(votes[i], phi)
    return votes




import os
import ast

from mapel.core.glossary import NICE_NAME


def import_real_instance(experiment, shift=False):
    """ Import real ordinal election form .soc file """

    votes = [0, 0]

    file_name = f'{experiment.instance_id}.mi'
    path = os.path.join(os.getcwd(), "election", experiment.experiment_id, "instances", file_name)
    with open(path, 'r') as my_file:

        params = 0
        first_line = my_file.readline()

        if first_line[0] != '#':
            model_id = 'empty'
            num_agents = int(first_line)
        else:
            first_line = first_line.strip().split()
            model_id = first_line[1]

            if len(first_line) <= 2:
                params = {}
            else:
                params = ast.literal_eval(" ".join(first_line[2:]))

            num_agents = int(my_file.readline())

        for s in range(2):

            if s == 1:
                num_agents = int(my_file.readline())

            num_candidates = num_agents

            for _ in range(num_agents):
                my_file.readline()

            line = my_file.readline().rstrip("\n").split(',')
            num_voters = int(line[0])
            num_options = int(line[2])
            votes[s] = [[0 for _ in range(num_candidates)] for _ in range(num_voters)]

            for j in range(num_options):
                line = my_file.readline().rstrip("\n").split(',')
                _id = int(line[0])

                for k in range(1):
                    for el in range(num_candidates):
                        votes[s][_id][el] = int(line[el + 1])

            if shift:
                for i in range(num_voters):
                    for j in range(num_candidates):
                        votes[s][i][j] -= 1

    rev_NICE_NAME = {v: k for k, v in NICE_NAME.items()}

    if model_id in rev_NICE_NAME:
        model_id = rev_NICE_NAME[model_id]

    return votes, num_agents, params, model_id


#!/usr/bin/env python

import mapel.marriages.features.experiments_marriage as exp_mar

# MAPPING #
def get_feature(feature_id):
    return {'summed_rank_maximal_matching': exp_mar.summed_rank_maximal_matching,
            'summed_rank_minimal_matching': exp_mar.summed_rank_minimal_matching,
            'minimal_rank_maximizing_matching': exp_mar.minimal_rank_maximizing_matching,
            'avg_num_of_bps_for_rand_matching': exp_mar.avg_number_of_bps_for_random_matching,
            'num_of_bps_min_weight': exp_mar.number_of_bps_maximumWeight,
            }.get(feature_id)



#!/usr/bin/env python


from time import time
from typing import Callable

import copy
import os
import csv
import numpy as np

from mapel.core.inner_distances import map_str_to_func
from mapel.core.objects.Experiment import Experiment
from mapel.marriages.distances import main_marriages_distances as mrd
from mapel.marriages.objects.Marriages import Marriages


def get_distance(election_1: Marriages, election_2: Marriages,
                 distance_id: str = None) -> (float, list):
    """ Return: distance between ordinal elections, (if applicable) optimal matching """
    inner_distance, main_distance = extract_distance_id(distance_id)

    metrics_without_params = {

    }

    metrics_with_inner_distance = {
        'mutual_attraction': mrd.compute_retrospective_distance,
        # 'positionwise': mrd.compute_positionwise_distance,
        # 'pos_swap': mrd.compute_pos_swap_distance,
        # 'swap_bf': mrd.compute_swap_bf_distance,
        # 'pairwise': mrd.compute_pairwise_distance,
    }

    if main_distance in metrics_without_params:
        return metrics_without_params.get(main_distance)(election_1, election_2)

    elif main_distance in metrics_with_inner_distance:
        return metrics_with_inner_distance.get(main_distance)(election_1, election_2,
                                                              inner_distance)


def extract_distance_id(distance_id: str) -> (Callable, str):
    if '-' in distance_id:
        inner_distance, main_distance = distance_id.split('-')
        inner_distance = map_str_to_func(inner_distance)
    else:
        main_distance = distance_id
        inner_distance = None
    return inner_distance, main_distance


def run_single_process(exp: Experiment, instances_ids: list,
                      distances: dict, times: dict, matchings: dict,
                      printing: bool) -> None:
    """ Single thread for computing distances """

    for instance_id_1, instance_id_2 in instances_ids:
        if printing:
            print(instance_id_1, instance_id_2)
        start_time = time()
        distance = get_distance(copy.deepcopy(exp.instances[instance_id_1]),
                                copy.deepcopy(exp.instances[instance_id_2]),
                                distance_id=copy.deepcopy(exp.distance_id))
        print(distance)
        if type(distance) is tuple:
            distance, matching = distance
            matching = np.array(matching)
            matchings[instance_id_1][instance_id_2] = matching
            matchings[instance_id_2][instance_id_1] = np.argsort(matching)
        distances[instance_id_1][instance_id_2] = distance
        distances[instance_id_2][instance_id_1] = distances[instance_id_1][instance_id_2]
        times[instance_id_1][instance_id_2] = time() - start_time
        times[instance_id_2][instance_id_1] = times[instance_id_1][instance_id_2]


def run_multiple_processes(exp: Experiment, instances_ids: list,
                      distances: dict, times: dict, matchings: dict,
                      printing: bool, t) -> None:
    """ Single thread for computing distances """

    for instance_id_1, instance_id_2 in instances_ids:
        if t == 0 and printing:
            print(instance_id_1, instance_id_2)
        start_time = time()
        distance = get_distance(copy.deepcopy(exp.instances[instance_id_1]),
                                copy.deepcopy(exp.instances[instance_id_2]),
                                distance_id=copy.deepcopy(exp.distance_id))
        if type(distance) is tuple:
            distance, matching = distance
            matching = np.array(matching)
            matchings[instance_id_1][instance_id_2] = matching
            matchings[instance_id_2][instance_id_1] = np.argsort(matching)
        distances[instance_id_1][instance_id_2] = distance
        distances[instance_id_2][instance_id_1] = distances[instance_id_1][instance_id_2]
        times[instance_id_1][instance_id_2] = time() - start_time
        times[instance_id_2][instance_id_1] = times[instance_id_1][instance_id_2]

    if exp.is_exported:
        _store_distances(exp, instances_ids, distances, times, t)


def _store_distances(exp, instances_ids, distances, times, t):
    """ Store distances to file """
    file_name = f'{exp.distance_id}_p{t}.csv'
    path = os.path.join(os.getcwd(), "election", exp.experiment_id, "distances", file_name)
    with open(path, 'w', newline='') as csv_file:
        writer = csv.writer(csv_file, delimiter=';')
        writer.writerow(["instance_id_1", "instance_id_2", "distance", "time"])
        for election_id_1, election_id_2 in instances_ids:
            distance = float(distances[election_id_1][election_id_2])
            time_ = float(times[election_id_1][election_id_2])
            writer.writerow([election_id_1, election_id_2, distance, time_])

# # # # # # # # # # # # # # # #
# LAST CLEANUP ON: 13.10.2021 #
# # # # # # # # # # # # # # # #




from .objects.MarriagesExperiment import MarriagesExperiment
from .objects.Marriages import Marriages
from . import cultures_ as cult
from .distances_ import get_distance


def prepare_online_marriages_experiment(**kwargs):
    return prepare_marriages_experiment(**kwargs, is_exported=False, is_imported=False)


def prepare_offline_marriages_experiment(**kwargs):
    return prepare_marriages_experiment(**kwargs, is_exported=True, is_imported=True)


def prepare_marriages_experiment(experiment_id=None,
                                 instance_type=None,
                                 distance_id=None,
                                 is_imported=None,
                                 is_shifted=False,
                                 is_exported=None,
                                 embedding_id=None):

    return MarriagesExperiment(experiment_id=experiment_id,
                               is_imported=is_imported,
                               distance_id=distance_id,
                               instance_type=instance_type,
                               embedding_id=embedding_id,
                               is_shifted=is_shifted,
                               is_exported=is_exported)


def generate_marriages_instance(**kwargs):
    instance = Marriages('virtual', 'tmp', **kwargs)
    instance.prepare_instance()
    return instance


def generate_marriages_votes(**kwargs):
    return cult.generate_votes(**kwargs)


def compute_distance(*args, **kwargs):
    return get_distance(*args, **kwargs)


import numpy as np


def generate_ic_votes(num_agents: int = None, **kwargs):

    return [list(np.random.permutation(num_agents)) for _ in range(num_agents)]


def generate_id_votes(num_agents: int = None, **kwargs):

    return [list(range(num_agents)) for _ in range(num_agents)]


def generate_asymmetric_votes(num_agents: int = None, **kwargs):
    votes = [list(range(num_agents)) for _ in range(num_agents)]
    votes_left = [rotate(vote, shift+1) for shift, vote in enumerate(votes)]
    votes = [list(range(num_agents)) for _ in range(num_agents)]
    votes_right = [rotate(vote, shift) for shift, vote in enumerate(votes)]
    return [votes_left, votes_right]


def generate_group_ic_votes(num_agents: int = None, proportion: int = 0.5, **kwargs):
    """ Impartial Culture with two groups """

    size_1 = int(proportion * num_agents)
    size_2 = int(num_agents - size_1)

    votes_1 = [list(np.random.permutation(size_1)) +
               list(np.random.permutation([j for j in range(size_1, num_agents)]))
               for _ in range(size_1)]

    votes_2 = [list(np.random.permutation([j for j in range(size_1, num_agents)])) +
               list(np.random.permutation(size_1))
               for _ in range(size_2)]

    votes = votes_1 + votes_2

    return votes


def generate_symmetric_votes(num_agents: int = None, **kwargs):

    num_rounds = num_agents - 1

    def next(agents):
        first = agents[0]
        last = agents[-1]
        middle = agents[1:-1]
        new_agents = [first, last]
        new_agents.extend(middle)
        return new_agents

    agents = [i for i in range(num_agents)]
    rounds = []

    for _ in range(num_rounds):
        pairs = []
        for i in range(num_agents // 2):
            agent_1 = agents[i]
            agent_2 = agents[num_agents - 1 - i]
            pairs.append([agent_1, agent_2])
        rounds.append(pairs)
        agents = next(agents)

    votes = np.zeros([num_agents, num_agents], dtype=int)

    for pos, partition in enumerate(rounds):
        for x, y in partition:
            votes[x][pos+1] = y
            votes[y][pos+1] = x

    for i in range(num_agents):
        votes[i][0] = i

    return votes

# HELPER
def rotate(vector, shift):
    shift = shift % len(vector)
    return vector[shift:] + vector[:shift]


# # # # # # # # # # # # # # # #
# LAST CLEANUP ON: 14.10.2021 #
# # # # # # # # # # # # # # # #



import gurobipy as gp
from gurobipy import GRB
from random import shuffle
import statistics
import warnings
import sys
import time
import networkx as nx

def generate_instance(num_agents):
    instance=[]
    for i in range(num_agents):
        pref = [x for x in range(num_agents) if x != i]
        shuffle(pref)
        instance.append(pref)
    return instance


#Only works for even number of agents
def swap_distance_to_stable(instance):
    num_agents=len(instance)
    m = gp.Model("mip1")
    m.setParam('OutputFlag', False)
    #who is matched to whom
    x = m.addVars(num_agents, num_agents, lb=0, ub=1, vtype=GRB.BINARY)
    #am I matched to rank i or better
    gm = m.addVars(num_agents, num_agents-1, lb=0, ub=1, vtype=GRB.BINARY)
    #by how much have my partner been swaped up in my preferences
    yy = m.addVars(num_agents, lb=0, ub=num_agents, vtype=GRB.INTEGER)

    for i in range(num_agents):
        yy[i].start=0

    opt = m.addVar(vtype=GRB.INTEGER, lb=0, ub=num_agents*num_agents)
    m.addConstr(gp.quicksum(yy[i] for i in range(num_agents)) == opt)
    for i in range(num_agents):
        m.addConstr(x[i, i]  == 0)

    for i in range(num_agents):
        for j in range(num_agents):
            m.addConstr(x[i, j]  == x[j, i])

    for i in range(num_agents):
        for j in range(num_agents):
            if i!=j:
                for t in range(num_agents-1):
                    m.addConstr(x[i, j]*instance[i].index(j)  <= t+yy[i]+num_agents*(1-gm[i,t]))


    for i in range(num_agents):
        m.addConstr(gp.quicksum(x[i, j] for j in range(num_agents)) == 1)

    for i in range(num_agents):
        for j in range(i+1,num_agents):
            m.addConstr(gm[i,instance[i].index(j)]+gm[j,instance[j].index(i)]>=1)
    m.setObjective(opt, GRB.MINIMIZE)
    m.optimize()

    return int(m.objVal)

def delete_distance_to_stable(instance):
    num_agents=len(instance)
    m = gp.Model("mip1")
    m.setParam('OutputFlag', False)
    x = m.addVars(num_agents, num_agents, lb=0, ub=1, vtype=GRB.BINARY)
    y = m.addVars(num_agents, lb=0, ub=1, vtype=GRB.BINARY)
    for i in range(num_agents):
        y[i].start=0
    opt = m.addVar(vtype=GRB.INTEGER, lb=0, ub=num_agents)
    m.addConstr(gp.quicksum(y[i] for i in range(num_agents)) == opt)
    for i in range(num_agents):
        m.addConstr(x[i, i]  == 0)
    for i in range(num_agents):
        for j in range(num_agents):
            m.addConstr(x[i, j]  == x[j, i])
    for i in range(num_agents):
        m.addConstr(gp.quicksum(x[i, j] for j in range(num_agents)) <= (1-y[i]))
    for i in range(num_agents):
        for j in range(i+1,num_agents):
            better_pairs=[]
            for t in range(0,instance[i].index(j)+1):
                better_pairs.append([i,instance[i][t]])
            for t in range(0,instance[j].index(i)+1):
                better_pairs.append([j,instance[j][t]])
            m.addConstr(gp.quicksum(x[a[0], a[1]] for a in better_pairs) >= 1-y[i]-y[j])

    m.setObjective(opt, GRB.MINIMIZE)
    m.optimize()
    matching={}
    for i in range(num_agents):
        for j in range(num_agents):
            if x[i, j].X == 1:
                matching[i]=j
                matching[j]=i
    return int(m.objVal)


def min_num_blocking_agents_matching(instance):
    num_agents=len(instance)
    m = gp.Model("mip1")
    m.setParam('OutputFlag', False)
    x = m.addVars(num_agents, num_agents, lb=0, ub=1, vtype=GRB.BINARY)
    y = m.addVars(num_agents, lb=0, ub=1, vtype=GRB.BINARY)
    for i in range(num_agents):
        y[i].start=0
    opt = m.addVar(vtype=GRB.INTEGER, lb=0, ub=num_agents)
    m.addConstr(gp.quicksum(y[i] for i in range(num_agents)) <= opt)
    for i in range(num_agents):
        m.addConstr(x[i, i]  == 0)
    for i in range(num_agents):
        for j in range(num_agents):
            m.addConstr(x[i, j]  == x[j, i])
    for i in range(num_agents):
        m.addConstr(gp.quicksum(x[i, j] for j in range(num_agents)) <= 1)
    for i in range(num_agents):
        for j in range(i+1,num_agents):
            better_pairs=[]
            for t in range(0,instance[i].index(j)+1):
                better_pairs.append([i,instance[i][t]])
            for t in range(0,instance[j].index(i)+1):
                better_pairs.append([j,instance[j][t]])
            m.addConstr(gp.quicksum(x[a[0], a[1]] for a in better_pairs) >= 1-y[i]-y[j])

    m.setObjective(opt, GRB.MINIMIZE)
    m.optimize()
    matching={}
    for i in range(num_agents):
        for j in range(num_agents):
            if x[i, j].X == 1:
                matching[i]=j
                matching[j]=i
    print(m.objVal)
    return int(m.objVal)

#for i in range(100):
#    instance=generate_instance(50)
#    print(min_num_blocking_agents_matching(instance))
#    print(swap_distance_to_stable(instance))
#    print(delete_distance_to_stable(instance))



import gurobipy as gp
from gurobipy import GRB
from random import shuffle
import statistics
import warnings
import sys
import time
import networkx as nx
sys.setrecursionlimit(10000)
# warnings.filterwarnings("error")

def generate_instance(num_agents):
    man=[]
    woman = []
    for i in range(num_agents):
        pref = [x for x in range(num_agents)]
        shuffle(pref)
        man.append(pref)
    for i in range(num_agents):
        pref = [x for x in range(num_agents)]
        shuffle(pref)
        woman.append(pref)
    return [man, woman]

def number_blockingPairs(instance,matching):
    #matching=[[3, 1, 0, 2, 4], [2, 1, 3, 0, 4]]
    bps=0
    num_agents=len(instance[0])
    for i in range(num_agents):
        for j in range(num_agents):
            #print(i,j)
            partner_i=matching[0][i]
            partneri_index=instance[0][i].index(partner_i)
            partner_j = matching[1][j]
            partnerj_index = instance[1][j].index(partner_j)
            if instance[0][i].index(j)<partneri_index:
                if instance[1][j].index(i) < partnerj_index:
                    #print(i,j)
                    bps+=1
    #print(bps)
    return bps




#Only call for instances that admit a stable matchig
#summed: Set to true we try to optimize the summed rank of agents for their partner in the matching, set to false we optimize the minimum rank
#Best (only relevant if summed is set to true): Set to true we output the best possible matching, set to false the worst one
def rank_matching(instance,best,summed):
    num_agents=len(instance[0])
    m = gp.Model("mip1")
    m.setParam('OutputFlag', False)
    x = m.addVars(num_agents, num_agents, lb=0, ub=1, vtype=GRB.BINARY)
    opt = m.addVar(vtype=GRB.INTEGER, lb=0, ub=num_agents*num_agents)
    opt1 = m.addVar(vtype=GRB.INTEGER, lb=0, ub=num_agents * num_agents)
    opt2 = m.addVar(vtype=GRB.INTEGER, lb=0, ub=num_agents * num_agents)
    for i in range(num_agents):
        m.addConstr(gp.quicksum(x[i, j] for j in range(num_agents)) <= 1)
        m.addConstr(gp.quicksum(x[j, i] for j in range(num_agents)) <= 1)
    for i in range(num_agents):
        for j in range(num_agents):
            #print(i,j)
            better_pairs=[]
            for t in range(0,instance[0][i].index(j)+1):
                better_pairs.append([i,instance[0][i][t]])
            for t in range(0,instance[1][j].index(i)+1):
                better_pairs.append([instance[1][j][t],j])
            #print(better_pairs)
            #exit()
            m.addConstr(gp.quicksum(x[a[0], a[1]] for a in better_pairs) >= 1)
    if summed:
        #print(instance[0])
        m.addConstr(gp.quicksum(
            instance[0][i].index(j) * x[i, j] for i in range(num_agents) for j in range(num_agents)) == opt1)
        m.addConstr(gp.quicksum(
            instance[1][i].index(j) * x[j, i] for i in range(num_agents) for j in range(num_agents)) == opt2)
        m.addConstr(opt1 + opt2 == opt)
        if best:
            m.setObjective(opt, GRB.MAXIMIZE)
        else:
            m.setObjective(opt, GRB.MINIMIZE)
    else:
        for i in range(num_agents):
            m.addConstr(gp.quicksum(instance[0][j].index(i) * x[j, i] for j in range(num_agents))<=opt)
            m.addConstr(gp.quicksum(instance[1][j].index(i) * x[i, j] for j in range(num_agents))<= opt)
        m.setObjective(opt, GRB.MINIMIZE)
    m.optimize()
    matching1={}
    matching2 = {}
    for i in range(num_agents):
        for j in range(num_agents):
            if x[i, j].X == 1:
                matching1[i]=j
                matching2[j]=i
    return int(m.objVal), [matching1,matching2]



def summed_rank_maximal_matching(instance):
    try:
        val,matching= rank_matching(instance,True,True)
    except:
        return None
    # if number_blockingPairs(instance,matching)>0:
    #     print('ERROR IN summed_rank_maximal_matching')
    #     exit(0)
    return val

def summed_rank_minimal_matching(instance):
    try:
        val,matching= rank_matching(instance,False,True)
    except:
        return None
    # if number_blockingPairs(instance,matching)>0:
    #     print('ERROR IN summed_rank_minimal_matching')
    #     exit(0)
    return val

def minimal_rank_maximizing_matching(instance):
    try:
        val,matching=rank_matching(instance,True,False)
    except:
        return None
    # if number_blockingPairs(instance,matching)>0:
    #     print('ERROR IN minimal_rank_maximizing_matching')
    #     exit(0)
    return val

def avg_number_of_bps_for_random_matching(instance,iterations=100):
    instance = instance.votes
    bps=[]
    num_agents=len(instance[0])
    for _ in range(iterations):
        #create random matching
        agents=list(range(num_agents))
        shuffle(agents)
        m2=[]
        for i in range(num_agents):
            m2.append(agents.index(i))
        #print([agents,m2])
        #exit()
        bps.append(number_blockingPairs(instance,[agents,m2]))
    return statistics.mean(bps), statistics.stdev(bps)

def number_of_bps_maximumWeight(instance):
    instance = instance.votes
    num_agents=len(instance[0])
    G = nx.Graph()
    for i in range(num_agents):
        for j in range(num_agents):
            G.add_edge(i,j+num_agents,weight=2*(num_agents-1)-instance[0][i].index(j)-instance[1][j].index(i))
            #G.add_edge(i, j, weight=instance[i].index(j) + instance[j].index(i))
    matching=nx.max_weight_matching(G, maxcardinality=True)
    #print(matching)
    #exit()
    matching_dict_m={}
    matching_dict_w = {}
    for p in matching:
        small=min(p[0],p[1])
        big = max(p[0], p[1])
        matching_dict_m[small]=big-num_agents
        matching_dict_w[big-num_agents] = small
    #print(matching_dict_m,matching_dict_w)
    return number_blockingPairs(instance,[matching_dict_m,matching_dict_w])

'''
j=0
for i in range(100):
    instance=generate_instance(50)
    #print('---------------------')
    #print(instance)
    #print(number_of_bps_maximumWeight(instance))
    print('NEW')
    print(summed_rank_maximal_matching(instance))
    print(summed_rank_minimal_matching(instance))
    print(minimal_rank_maximizing_matching(instance))
    #print(average_number_of_bps_for_random_matching(instance))
    #swap_distance_to_stable(instance)
    #delete_distance_to_stable(instance)
    #min_num_blocking_agents_matching(instance)
    #print(number_of_bps_maximumWeight(instance))
    #print(min_num_bps_matching(instance))


exit()
instance1=generate_instance(10)
instance2=generate_instance(10)
start = time.process_time()
print(spear_distance_linear(instance1,instance2))
print(time.process_time() - start)
exit()
'''


#!/usr/bin/env python
import ast
import csv
import itertools
import os
from multiprocessing import Process
from time import sleep

from mapel.core.objects.Experiment import Experiment
from mapel.marriages.objects.MarriagesFamily import MarriagesFamily
from mapel.marriages.objects.Marriages import Marriages
import mapel.marriages.distances as metr
import mapel.marriages.features.basic_features as basic
import mapel.marriages.features as features
from mapel.core.persistence.experiment_imports import get_values_from_csv_file
from mapel.core.utils import make_folder_if_do_not_exist
import mapel.core.persistence.experiment_exports as exports

try:
    from sklearn.manifold import MDS
    from sklearn.manifold import TSNE
    from sklearn.manifold import SpectralEmbedding
    from sklearn.manifold import LocallyLinearEmbedding
    from sklearn.manifold import Isomap
except ImportError as error:
    MDS = None
    TSNE = None
    SpectralEmbedding = None
    LocallyLinearEmbedding = None
    Isomap = None
    print(error)


class MarriagesExperiment(Experiment):
    """Abstract set of elections."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.default_num_agents = 10

        self.matchings = {}

        # election.import_matchings()

    def import_matchings(self):
        matchings = {}

        path = os.path.join(os.getcwd(), 'election', self.experiment_id, 'features',
                            'stable_sr.csv')
        with open(path, 'r', newline='') as csv_file:
            reader = csv.DictReader(csv_file, delimiter=';')

            for row in reader:
                election_id = row['instance_id']
                value = row['matching']
                if value == '':
                    matchings[election_id] = None
                else:
                    matchings[election_id] = value

        self.matchings = matchings

    def add_instances_to_experiment(self):

        instances = {}

        for family_id in self.families:

            ids = []
            if self.families[family_id].single:
                election_id = family_id
                election = Marriages(self.experiment_id, election_id)
                instances[election_id] = election
                ids.append(str(election_id))
            else:
                for j in range(self.families[family_id].size):
                    election_id = family_id + '_' + str(j)
                    election = Marriages(self.experiment_id, election_id)
                    instances[election_id] = election
                    ids.append(str(election_id))

            self.families[family_id].instance_ids = ids

        return instances

    def compute_distances(self, distance_id: str = 'emd-positionwise', num_processes: int = 1,
                          self_distances: bool = False, vector_type: str = 'A',
                          printing: bool = False) -> None:
        """ Compute distances between elections (using processes) """

        self.distance_id = distance_id

        if '-pairwise' in distance_id:
            for instance in self.instances.values():
                instance.votes_to_pairwise_matrix()

        matchings = {election_id: {} for election_id in self.instances}
        distances = {election_id: {} for election_id in self.instances}
        times = {election_id: {} for election_id in self.instances}

        ids = []
        for i, election_1 in enumerate(self.instances):
            for j, election_2 in enumerate(self.instances):
                if i == j:
                    if self_distances:
                        ids.append((election_1, election_2))
                elif i < j:
                    ids.append((election_1, election_2))

        num_distances = len(ids)

        if self.experiment_id == 'virtual' or num_processes == 1:
            metr.run_single_process(self, ids, distances, times, matchings, printing)

        else:
            processes = []
            for t in range(num_processes):
                print(f'Starting thread: {t}')
                sleep(0.1)
                start = int(t * num_distances / num_processes)
                stop = int((t + 1) * num_distances / num_processes)
                instances_ids = ids[start:stop]

                process = Process(target=metr.run_multiple_processes, args=(self, instances_ids,
                                                                            distances, times,
                                                                            matchings,
                                                                            printing, t))
                process.start()
                processes.append(process)

            for process in processes:
                process.join()

            distances = {instance_id: {} for instance_id in self.instances}
            times = {instance_id: {} for instance_id in self.instances}
            for t in range(num_processes):

                file_name = f'{distance_id}_p{t}.csv'
                path = os.path.join(os.getcwd(), "election", self.experiment_id, "distances",
                                    file_name)

                with open(path, 'r', newline='') as csv_file:
                    reader = csv.DictReader(csv_file, delimiter=';')

                    for row in reader:
                        distances[row['instance_id_1']][row['instance_id_2']] = float(
                            row['distance'])
                        times[row['instance_id_1']][row['instance_id_2']] = float(row['time'])

        if self.is_exported:
            exports.export_distances_to_file(self, distance_id, distances, times)

        self.distances = distances
        self.times = times
        self.matchings = matchings

    def import_controllers(self):
        """ Import controllers from a file """

        families = {}

        path = os.path.join(os.getcwd(), 'election', self.experiment_id, 'map.csv')
        file_ = open(path, 'r')

        header = [h.strip() for h in file_.readline().split(';')]
        reader = csv.DictReader(file_, fieldnames=header, delimiter=';')

        starting_from = 0
        for row in reader:

            model_id = None
            color = None
            label = None
            params = None
            alpha = None
            size = None
            marker = None
            num_agents = None
            family_id = None
            show = True


            if 'culture_id' in row.keys():
                model_id = str(row['culture_id']).strip()

            if 'color' in row.keys():
                color = str(row['color']).strip()

            if 'label' in row.keys():
                label = str(row['label'])

            if 'family_id' in row.keys():
                family_id = str(row['family_id'])

            if 'params' in row.keys():
                params = ast.literal_eval(str(row['params']))

            if 'alpha' in row.keys():
                alpha = float(row['alpha'])

            if 'size' in row.keys():
                size = int(row['size'])

            if 'marker' in row.keys():
                marker = str(row['marker']).strip()

            if 'num_agents' in row.keys():
                num_agents = int(row['num_agents'])

            if 'path' in row.keys():
                path = ast.literal_eval(str(row['path']))

            if 'show' in row.keys():
                show = row['show'].strip() == 'process_id'

            single_instance = size == 1

            families[family_id] = MarriagesFamily(model_id=model_id,
                                                  family_id=family_id,
                                                  params=params, label=label,
                                                  color=color, alpha=alpha, show=show,
                                                  size=size, marker=marker,
                                                  starting_from=starting_from,
                                                  num_agents=num_agents, path=path,
                                                  single=single_instance)
            starting_from += size

        self.num_families = len(families)
        self.num_instances = sum([families[family_id].size for family_id in families])
        self.main_order = [i for i in range(self.num_instances)]

        file_.close()
        return families

    def prepare_instances(self):

        if self.instances is None:
            self.instances = {}

        for family_id in self.families:

            new_instances = self.families[family_id].prepare_family(
                is_exported=self.is_exported,
                experiment_id=self.experiment_id)

            for instance_id in new_instances:
                self.instances[instance_id] = new_instances[instance_id]

    def compute_stable_sr(self):
        for instance_id in self.instances:
            print(instance_id)
            if instance_id in ['roommates_test']:
                self.matchings[instance_id] = 'None'
            else:
                usable_matching = basic.compute_stable_SR(self.instances[instance_id].votes)
                self.matchings[instance_id] = usable_matching

        if self.is_exported:

            path_to_folder = os.path.join(os.getcwd(), "election", self.experiment_id,
                                          "features")
            make_folder_if_do_not_exist(path_to_folder)
            path_to_file = os.path.join(path_to_folder, f'stable_sr.csv')

            with open(path_to_file, 'w', newline='') as csv_file:
                writer = csv.writer(csv_file, delimiter=';')
                writer.writerow(
                    ["instance_id", "matching"])

                for instance_id in self.instances:
                    usable_matching = self.matchings[instance_id]
                    writer.writerow([instance_id, usable_matching])

    def compute_feature(self, feature_id: str = None, feature_params=None) -> dict:

        if feature_params is None:
            feature_params = {}

        feature_dict = {'value': {}, 'time': {}, 'std': {}}

        features_with_time = {}
        features_with_std = {'avg_num_of_bps_for_rand_matching',
                             'avg_number_of_bps_for_random_matching'}

        # print(election.matchings)

        if feature_id == 'summed_rank_difference':
            minimal = get_values_from_csv_file(self, feature_id='summed_rank_minimal_matching')
            maximal = get_values_from_csv_file(self, feature_id='summed_rank_maximal_matching')

            for instance_id in self.instances:
                if minimal[instance_id] is None:
                    value = 'None'
                else:
                    value = abs(maximal[instance_id] - minimal[instance_id])
                feature_dict['value'][instance_id] = value
                feature_dict['time'][instance_id] = 0

        else:

            for instance_id in self.instances:
                print(instance_id)
                feature = features.get_feature(feature_id)
                instance = self.instances[instance_id]
                # if feature_id in ['summed_rank_minimal_matching'] \
                #         and election.matchings[instance_id] is None:
                #     value = 'None'
                # else:
                value = feature(instance.votes)
                print(value)
                #
                # elif feature_id in ['largest_cohesive_group', 'number_of_cohesive_groups',
                #                     'number_of_cohesive_groups_brute',
                #                     'proportionality_degree_pav',
                #                     'proportionality_degree_av',
                #                     'proportionality_degree_cc',
                #                     'justified_ratio',
                #                     'cohesiveness',
                #                     'partylist',
                #                     'highest_cc_score',
                #                     'highest_hb_score']:
                #     value = feature(election, feature_params)
                #
                # elif feature_id in {'avg_distortion_from_guardians',
                #                     'worst_distortion_from_guardians',
                #                     'distortion_from_all',
                #                     'distortion_from_top_100'}:
                #     value = feature(election, election_id)
                # else:
                #     value = feature(election)

                if feature_id in features_with_time:
                    feature_dict['value'][instance_id] = value[0]
                    feature_dict['time'][instance_id] = value[1]
                elif feature_id in features_with_std:
                    feature_dict['value'][instance_id] = value[0]
                    feature_dict['std'][instance_id] = value[1]
                else:
                    feature_dict['value'][instance_id] = value

        if self.is_exported:

            path_to_folder = os.path.join(os.getcwd(), "election", self.experiment_id,
                                          "features")
            make_folder_if_do_not_exist(path_to_folder)
            path_to_file = os.path.join(path_to_folder, f'{feature_id}.csv')

            with open(path_to_file, 'w', newline='') as csv_file:
                writer = csv.writer(csv_file, delimiter=';')

                if feature_id in features_with_time:
                    writer.writerow(["instance_id", "value", 'time'])
                    for key in feature_dict['value']:
                        writer.writerow([key, feature_dict['value'][key], round(feature_dict['time'][key],3)])
                elif feature_id in features_with_std:
                    writer.writerow(["instance_id", "value", 'std'])
                    for key in feature_dict['value']:
                        writer.writerow([key, feature_dict['value'][key], round(feature_dict['std'][key],3)])
                else:
                    writer.writerow(["instance_id", "value"])
                    for key in feature_dict['value']:
                        writer.writerow([key, feature_dict['value'][key]])

        self.features[feature_id] = feature_dict
        return feature_dict

    def create_structure(self) -> None:

        dirs = ["experiments", "images", "trash"]
        for dir in dirs:
            if not os.path.isdir(dir):
                os.mkdir(os.path.join(os.getcwd(), dir))

        try:
            os.mkdir(os.path.join(os.getcwd(), "election", self.experiment_id))
            os.mkdir(os.path.join(os.getcwd(), "election", self.experiment_id, "distances"))
            os.mkdir(os.path.join(os.getcwd(), "election", self.experiment_id, "features"))
            os.mkdir(os.path.join(os.getcwd(), "election", self.experiment_id, "coordinates"))
            os.mkdir(os.path.join(os.getcwd(), "election", self.experiment_id, "instances"))
            os.mkdir(os.path.join(os.getcwd(), "election", self.experiment_id, "matrices"))

            # PREPARE MAP.CSV FILE
            path = os.path.join(os.getcwd(), "election", self.experiment_id, "map.csv")

            with open(path, 'w') as file_csv:
                file_csv.write("size;num_agents;culture_id;params;color;alpha;family_id;label;marker;show\n")
                file_csv.write("10;16;ic;{};black;1;IC;IC;o;process_id\n")
        except:
            pass



from mapel.core.glossary import *
from mapel.core.utils import *
from collections import Counter


def export_instance_to_a_file(experiment):
    """ Store votes in a file """

    path_to_folder = os.path.join(os.getcwd(), "election", experiment.experiment_id, "instances")
    make_folder_if_do_not_exist(path_to_folder)
    path_to_file = os.path.join(path_to_folder, f'{experiment.instance_id}.mi')

    with open(path_to_file, 'w') as file_:

        if experiment.model_id in NICE_NAME:
            file_.write("# " + NICE_NAME[experiment.model_id] + " " + str(experiment.params) + "\n")
        else:
            file_.write("# " + experiment.model_id + " " + str(experiment.params) + "\n")

        for s in range(2):

            file_.write(str(experiment.num_agents) + "\n")

            for i in range(experiment.num_agents):
                if s == 0:
                    file_.write(str(i) + ', a' + str(i) + "\n")
                else:
                    file_.write(str(i) + ', b' + str(i) + "\n")

            votes = experiment.votes[s]

            file_.write(str(experiment.num_agents) + ', ' + str(experiment.num_agents) + ', ' +
                        str(len(votes)) + "\n")

            for i in range(len(votes)):
                file_.write(str(i) + ', ')
                for j in range(experiment.num_agents):
                    file_.write(str(int(votes[i][j])))
                    if j < len(votes[i]) - 1:
                        file_.write(", ")
                file_.write("\n")




def convert(votes):
    new_votes = []
    for j in range(len(votes)):
        vote = list(votes[j])
        vote.remove(j)
        new_votes.append(vote)
    return new_votes


def remove_first(votes):
    new_votes = []
    for j in range(len(votes)):
        vote = list(votes[j][1:])
        new_votes.append(vote)
    return new_votes
